#include <netdb.h>  /* gethostbyname */
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include "statetransferd.h"

/* host can be a hostname or dotted-quad IP */
int in_addr_for_host(struct in_addr *addr, const char *host)
{
	struct hostent *hent;

	//fprintf(stderr, "%s: hostname: %s\n", __FUNCTION__, host);
	if ((hent = gethostbyname(host)) == NULL) {
		fprintf(stderr, "Invalid hostname or IP address: %s\n", host);
		return 1;
	}
	memcpy(addr, hent->h_addr, hent->h_length);

	return 0;
}

/* caller must free */
char * dotted_quad_for_long(unsigned long netip)
{
	struct in_addr addr;
	char *quad = NULL, *s;

	addr.s_addr = netip;
	if ((s = inet_ntoa(addr)) == NULL) {
		fprintf(stderr, "Unable to make a dotted-quad from long %lx.\n", netip);
		goto out;
	}
	if ((quad = malloc(MAX_IP4_STRLEN)) == NULL) {
		fprintf(stderr, "Unable to malloc.\n");
		goto out;
	}
	memset(quad, '\0', MAX_IP4_STRLEN);
	memcpy(quad, s, strlen(s));
out:
	return quad;
}

/* caller must free */
char * dotted_quad_for_host(const char *host)
{
	struct in_addr addr;

	if (in_addr_for_host(&addr, host))
		return NULL;

	return dotted_quad_for_long(addr.s_addr);
}

/* returns LONG_MIN on error */
time_t
extract_boottime(const char *timefile)
{
	FILE *fp;
	char epoch[EPOCH_STRLEN];
	time_t t = LONG_MIN;

	if ((fp = fopen(timefile, "r")) == NULL) {
		syslog(LOG_INFO, "%s: couldn't open %s: %s\n", __FUNCTION__, timefile, strerror(errno));
		goto out;
	}
	if (fgets(epoch, EPOCH_STRLEN, fp) == NULL) {
		syslog(LOG_INFO, "%s: couldn't read epoch from %s: %s\n", __FUNCTION__, timefile, strerror(errno));
		goto out;
	}
	syslog(LOG_INFO, "%s: got existing '%s' boottime as %s\n", __FUNCTION__, timefile, epoch);
	t = (time_t)strtol(epoch, NULL, 10);
	if (t == 0 || t == LONG_MAX || t == LONG_MIN) {
		syslog(LOG_INFO, "%s: failed converting epoch string '%s' to time_t %ld\n", __FUNCTION__, epoch, (long)t);
		t = LONG_MIN;
	}
out:
	if (fp && fclose(fp)) {
		syslog(LOG_INFO, "%s: failed closing %s: %s\n", __FUNCTION__, timefile, strerror(errno));
		t = LONG_MIN;
	}

	return t;
}

/* returns LONG_MIN on error */
time_t
extract_client_boottime(const char *client_recdir)
{
	char tfile[CLT_TIMEFILE_PATHLEN];

	if (sprintf(tfile, "%s/%s", client_recdir, CLT_TIMEFILE) < 0) {
		syslog(LOG_INFO, "%s: failed building client timefile path: %s\n", __FUNCTION__, strerror(errno));
		return LONG_MIN;
	}
	return extract_boottime(tfile);
}

/*
 * - assumes you are in 'nfsd_recovery_dir'
 * - returns the previous boottime, or LONG_MIN on error
 */
time_t
set_server_boottime(time_t cur)
{
	FILE *fp;
	time_t prev;

	if ((prev = extract_boottime(SVR_CUR_TIMEFILE)) == LONG_MIN)
		syslog(LOG_INFO, "%s: couldn't extract server boottime.\n", __FUNCTION__);
	else if (rename(SVR_CUR_TIMEFILE, SVR_PREV_TIMEFILE))
		syslog(LOG_INFO, "%s: couldn't rename %s: %s\n", __FUNCTION__, SVR_CUR_TIMEFILE, strerror(errno));

	if ((fp = fopen(SVR_CUR_TIMEFILE, "w")) == NULL) {
		syslog(LOG_INFO, "%s: couldn't open new %s: %s\n", __FUNCTION__, SVR_CUR_TIMEFILE, strerror(errno));
		goto failed;
	}
	if ((fprintf(fp, "%ld", (long)cur)) < 0) {
		syslog(LOG_INFO, "%s: couldn't write new server boottime: %s\n", __FUNCTION__, strerror(errno));
		goto failed;
	}
out:
	if (fp && fclose(fp)) {
		syslog(LOG_INFO, "%s: failed closing %s: %s\n", __FUNCTION__, SVR_CUR_TIMEFILE, strerror(errno));
		prev = LONG_MIN;
	}
	return prev;
failed:
	prev = LONG_MIN;
	goto out;
}
