/*
 * statetransferd:	NFSv4 server state upcall daemon
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <rpc/pmap_clnt.h>

#include "statetransferd.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

static char *program;

/* XXX: extern these guys in statetransferd.h */
char *nfsd_pipefs_dir = NFSD_PIPEFS_DIR;
char *s_trans_pipe_path = S_TRANS_PIPE_PATH;
char *nfsd_recovery_dir = NFSD_RECOVERY_DIR;

int nfsd_pipefs_fd = -1;
int s_trans_pipe_fd = -1;
int saved_errno = 0;


void
free_dirents(struct dirent **ents, int nents)
{
	int i;

	for (i = 0; i < nents; i++)
		free(ents[i]);
	free(ents);
}

int
rm_rf(const char *dir)
{
	char *file;
	int i, nsubents;
	struct dirent **subents;

	if (!dir || strlen(dir) != HEXDIR_LEN) {
		syslog(LOG_INFO, "%s: '%s' does not appear to be a recovery directory name.\n", __FUNCTION__, dir);
		return -1;
	}
	chdir(dir);
	nsubents = scandir(".", &subents, NULL, alphasort);
	for (i = 0; i < nsubents; i++) {
		file = subents[i]->d_name;
		if (!strcmp(file, ".") || !strcmp(file, ".."))
			continue;
                if (unlink(file)) {
                    syslog(LOG_INFO, "%s: unable to unlink '%s' from within recovery dir '%s'.\n", __FUNCTION__, file, dir);
                    continue;
                }
	}
	chdir("..");
	free_dirents(subents, nsubents);

	if (rmdir(dir)) {
		syslog(LOG_INFO, "%s: rmdir %s error %s\n", __FUNCTION__, dir, strerror(errno));
		return -1;
	} else
		syslog(LOG_INFO, "%s: purged recovery dir '%s'.\n", __FUNCTION__, dir);

	return 0;
}

static void
s_trans_error(struct s_trans_msg *msg)
{
	syslog(LOG_INFO, "s_trans_error");
	msg->sm_len = 0;

	if (atomicio(write, s_trans_pipe_fd, msg, sizeof(*msg)) != sizeof(*msg)) {
		syslog(LOG_INFO, "s_trans_error: write failed! %s",
						strerror(errno));
	}
	syslog(LOG_INFO, "s_trans_error DONE");
}

/*
 * Create a buffer for writing
 */
static char *
s_trans_set_wbuf(struct s_trans_msg *stm, char *data, int datalen)
{
	char *buf;
	u_int32_t *p;

	syslog(LOG_INFO, "s_trans_set_wbuf: len %lu\n", sizeof(*stm) + datalen);

	/* XXX need to limit datalen */
	if ((buf = calloc(1, sizeof(*stm) + datalen)) == NULL)
		return buf;

	p = (u_int32_t *)buf;
	*p++ = S_TRANS_STATUS_SUCCESS;
	*p++ = stm->sm_type;
	*p++ = stm->sm_origin;
	*p++ = datalen;

	if (data && datalen > 0) {
		memcpy((char *)p, data, datalen);
	}
	return buf;
}

static u_int32_t *
s_trans_encode_buf(u_int32_t *p, char *ptr, int nbytes)
{
	*(p + XDR_QUADLEN(nbytes) -1) = 0;
	memcpy(p, ptr, nbytes);
	p += XDR_QUADLEN(nbytes);
	return p;
}

/*
 * returns the size (in bytes) of the XDR buffer needed to store any
 * remaining valid client recovery directory names, or -1 on error.
 */
int
filter_client_dirs(time_t server_boot, time_t prev_boot, int purge)
{
	int nents, i, nu32s = 0;
	time_t clt_boot;
	struct dirent **ents;
	char *ent;
	struct stat st;

	nents = scandir(nfsd_recovery_dir, &ents, NULL, alphasort);
	if (nents == -1) {
		syslog(LOG_INFO, "%s: scandir failed. %s\n", __FUNCTION__, strerror(errno));
		return -1;
	}

	syslog(LOG_INFO, "%s: srv boot: %ld  prev boot: %ld\n", __FUNCTION__, server_boot, prev_boot);
	/*
	 * MD5 ASCII client ID directory names are always 32 bytes use
	 * XDR_QUADLEN. add space for length.
	 */
	for (i = 0; i < nents; i++) {
		ent = ents[i]->d_name;
		syslog(LOG_INFO, "%s: ents[i]->d-name %s reclen %d, strlen %d\n", __FUNCTION__, ent, ents[i]->d_reclen, (int)strlen(ent));
		if (strlen(ent) == HEXDIR_LEN) {
			if (stat(ent, &st) || !(st.st_mode & S_IFDIR))
				continue;
			if (purge) {
				syslog(LOG_INFO, "%s: purging client %s\n", __FUNCTION__, ent);
				rm_rf(ent);
				continue;
			}
			if ((clt_boot = extract_client_boottime(ent)) == LONG_MIN) {
				syslog(LOG_INFO, "%s: failed extracting client boottime from dir %s -- purging\n", __FUNCTION__, ent);
				rm_rf(ent);
				continue;
			}
			if (clt_boot < prev_boot) {
				syslog(LOG_INFO, "%s: purging too-old (%ld vs %ld) client %s\n", __FUNCTION__, clt_boot, prev_boot, ent);
				rm_rf(ent);
			} else if (clt_boot > server_boot) {
				syslog(LOG_INFO, "%s: WARNING: got a client boot -later- than the server boot?? (%ld vs %ld)  TOLERATING and allowing reclaim\n", __FUNCTION__, clt_boot, server_boot);
				nu32s += XDR_QUADLEN(HEXDIR_LEN_PADDED) + 1;  /* counts u32s  XXX dmr remove */
			} else {
				syslog(LOG_INFO, "%s: client %s should be able to reclaim (%ld < %ld < %ld)\n", __FUNCTION__, ent, prev_boot, clt_boot, server_boot);
				nu32s += XDR_QUADLEN(HEXDIR_LEN_PADDED) + 1;  /* counts u32s */
			}
		}
	}
	free_dirents(ents, nents);
	
	/* convert from u32s to bytes */	
	return nu32s << 2;
}

/*
 * XXX need to limit write buffer and add readdir cookie to header
 * so that multiple upcalls can be used to get > write buffer data.
 * returns 0 on success, or -1 on error. 
 */
int
s_trans_load_state(struct s_trans_msg *stmp, char **wbufp, int *wbuflen)
{
	int nents, i, mlen, nbytes = 0, purge = 0, err = -1;
	uint32_t *p;
	time_t server_boot, prev_boot;
	struct dirent **ents;
	char *wbuf;

	if (stmp->sm_origin == S_TRANS_KNFSD) {
		/* if our knfsd is loading state, we need to update its boottime files. */
		/* skip past header; followed by server boot time (as epoch) */
		p = (uint32_t *)stmp;
		p += 4;
		server_boot = (time_t)*p;

		if ((prev_boot = set_server_boottime(server_boot)) == LONG_MIN) {
			syslog(LOG_INFO, "%s: WARNING: failed to get previous server boottime; will clear _all_ client recovery state.\n", __FUNCTION__);
			purge = 1;
		}
		syslog(LOG_INFO, "%s: starting LOAD_STATE from knfsd -- using cur %ld  prev %ld\n", __FUNCTION__, server_boot, prev_boot);
	} else if (stmp->sm_origin == S_TRANS_RPC) {
		/* if we got a load state RPC, we need to get our existing boottimes so we can
		 * filter properly. */
		server_boot = extract_boottime(SVR_CUR_TIMEFILE);
		prev_boot = extract_boottime(SVR_PREV_TIMEFILE);
		syslog(LOG_INFO, "%s: starting LOAD_STATE from RPC -- using cur %ld  prev %ld\n", __FUNCTION__, server_boot, prev_boot);
	} else {
		syslog(LOG_INFO, "%s: ERROR -- LOAD_STATE with sm_origin == %d ??\n", __FUNCTION__, stmp->sm_origin);
		goto out;
	}

	if ((nbytes = filter_client_dirs(server_boot, prev_boot, purge)) == -1) {
		syslog(LOG_INFO, "%s: ERROR: failed to load client recovery directory information.\n", __FUNCTION__);
		goto out;
	}

	mlen = sizeof(struct s_trans_msg) + nbytes;
	syslog(LOG_INFO, "%s: ALLOC wbuf of: %d bytes (%d + %d)\n", __FUNCTION__, mlen, (int)sizeof(struct s_trans_msg), nbytes);
	if ((wbuf = calloc(1, mlen)) == NULL) {
		syslog(LOG_INFO, "%s: ERROR: unable to alloc write buf.\n", __FUNCTION__);
		goto out;
	}

	/* header */
	p = (uint32_t *)wbuf;
	*p++ = S_TRANS_STATUS_SUCCESS;
	*p++ = S_TRANS_LOAD_REBOOT_STATE;
	*p++ = stmp->sm_origin;
	*p++ = nbytes; /* of XDR encoded data */

	/* XDR encoding the data */
	nents = scandir(nfsd_recovery_dir, &ents, NULL, alphasort);
	if (nents == -1) {
		syslog(LOG_INFO, "%s: 2nd scandir failed: %s\n", __FUNCTION__, strerror(errno));
		free(wbuf);
		*wbufp = NULL;
		*wbuflen = 0;
		goto out;
	}
	for (i = 0; i < nents; i++) {
		if (strlen(ents[i]->d_name) == HEXDIR_LEN) {
			syslog(LOG_INFO, "%s: encoding buf for eligible-to-recover %s\n", __FUNCTION__, ents[i]->d_name);
			*p++ = htonl(XDR_QUADLEN(HEXDIR_LEN_PADDED));
			p = s_trans_encode_buf(p, ents[i]->d_name, HEXDIR_LEN_PADDED);
		}
	}
	*wbufp = wbuf;
	*wbuflen = mlen;
	free_dirents(ents, nents);
	err = 0;

out:
	syslog(LOG_INFO, "%s: returns %d\n", __FUNCTION__, err);
	return err;
}

int
s_trans_add_state(struct s_trans_msg *stmp, char **wbufp, int *wbuflen)
{
	int err = -1;
	char dirname[HEXDIR_LEN + 1];
	char timefile[CLT_TIMEFILE_PATHLEN], ipfile[CLT_IPFILE_PATHLEN], *wbuf;
	time_t epoch;
	FILE *fp;
	struct stat st;
	uint32_t *p;
	struct in_addr iaddr;

	/* dmr XXX remove: this'll go away once minimal/upcall ADD_STATE has epoch, too */
	if (stmp->sm_origin == S_TRANS_RPC) {
		if (stmp->sm_len != (HEXDIR_LEN_PADDED + 2*sizeof(uint32_t))) {
			syslog(LOG_INFO, "%s: Invalid message data length\n", __FUNCTION__);
			stmp->sm_status = S_TRANS_STATUS_INVALIDMSG;
			goto out;
		}
		p = (uint32_t *)stmp;
		p += 4;
		memcpy(dirname, (char *)p, HEXDIR_LEN);
		dirname[HEXDIR_LEN] = '\0';
		syslog(LOG_INFO, "%s: RPC ADD_STATE dir saved as %s\n", __FUNCTION__, dirname);
		
		p += (HEXDIR_LEN_PADDED / 4);
		iaddr.s_addr = *p;
		syslog(LOG_INFO, "%s: RPC ADD_STATE (%d) (%s) for IP '%s'\n", __FUNCTION__, stmp->sm_type, dirname, inet_ntoa(iaddr));

		p += 1;
		epoch = *((time_t *)p);
		syslog(LOG_INFO, "%s: RPC ADD_STATE for IP '%s' epoch '%ld'\n", __FUNCTION__, inet_ntoa(iaddr), epoch);

		/* make the client recovery directory */
		if (mkdir(dirname, 0666)) {
			if (errno != EEXIST) {
				syslog(LOG_INFO, "%s: RPC failed to make dir '%s' for ADD STATE: %s\n", __FUNCTION__, dirname, strerror(errno));
				goto out;
			}
			if (stat(dirname, &st) || !(st.st_mode & S_IFDIR)) {
				syslog(LOG_INFO, "%s: RPC failed to make dir '%s' because a non-directory already has that name.\n", __FUNCTION__, dirname);
				goto out;
			}
		}
	} else {
		if (stmp->sm_len != (HEXDIR_LEN_PADDED + sizeof(uint32_t))) {
			syslog(LOG_INFO, "%s: Invalid message data length\n", __FUNCTION__);
			stmp->sm_status = S_TRANS_STATUS_INVALIDMSG;
			goto out;
		}
	
		/* skip past header; followed by a hexdir string and a u32 client ip */
		p = (uint32_t *)stmp;
		p += 4;
		memcpy(dirname, (char *)p, HEXDIR_LEN);
		dirname[HEXDIR_LEN] = '\0';
		syslog(LOG_INFO, "%s: ADD_STATE dir saved as %s\n", __FUNCTION__, dirname);
		iaddr.s_addr = *(p + (HEXDIR_LEN_PADDED / 4));
		syslog(LOG_INFO, "%s: just got ADD_STATE (%d) (%s) for IP '%s'\n", __FUNCTION__, stmp->sm_type, dirname, inet_ntoa(iaddr));
	
		/* make the client recovery directory */
		if (mkdir(dirname, 0666)) {
			if (errno != EEXIST) {
				syslog(LOG_INFO, "%s: failed to make dir '%s' for ADD STATE: %s\n", __FUNCTION__, dirname, strerror(errno));
				goto out;
			}
			if (stat(dirname, &st) || !(st.st_mode & S_IFDIR)) {
				syslog(LOG_INFO, "%s: failed to make dir '%s' because a non-directory already has that name.\n", __FUNCTION__, dirname);
				goto out;
			}
		}
	
		/* make the client boot timefile */
		epoch = time(NULL);
		if (epoch == (time_t)-1) {
			syslog(LOG_INFO, "%s: failed to get client timestamp.\n", __FUNCTION__);
			goto out;
		}
	}

	if (sprintf(timefile, "%s/%s", dirname, CLT_TIMEFILE) < 0) {
		syslog(LOG_INFO, "%s: failed to make client timestamp filename.\n", __FUNCTION__);
		goto out;
	}
	syslog(LOG_INFO, "%s: built client timestamp filename '%s' (dirname '%s')\n", __FUNCTION__, timefile, dirname);
	if ((fp = fopen(timefile, "w")) == NULL) {
		syslog(LOG_INFO, "%s: failed to create client timestamp file '%s'.\n", __FUNCTION__, timefile);
		/* XXX: rmdir the directory? */
		goto out;
	}
	if ((fprintf(fp, "%ld", epoch)) < 0) {
		syslog(LOG_INFO, "%s: failed to write epoch '%ld' to client timestamp file '%s'.\n", __FUNCTION__, epoch, timefile);
		/* XXX: rmdir the directory? */
		goto out;
	}
	fclose(fp);

	/* make the client ip file */
	if (sprintf(ipfile, "%s/%s", dirname, CLT_IPFILE) < 0) {
		syslog(LOG_INFO, "%s: failed to make client ip filename.\n", __FUNCTION__);
		goto out;
	}
	syslog(LOG_INFO, "%s: built client ip filename '%s' (dirname '%s')\n", __FUNCTION__, ipfile, dirname);
	if ((fp = fopen(ipfile, "w")) == NULL) {
		syslog(LOG_INFO, "%s: couldn't write client IP file.\n", __FUNCTION__);
		goto out;
	}
	if (fprintf(fp, "%s", inet_ntoa(iaddr)) < 0) {
		syslog(LOG_INFO, "%s: couldn't write IP %s to client IP file.\n", __FUNCTION__, inet_ntoa(iaddr));
		goto out;
	}
	fclose(fp);
	err = 0;

	/* XXX: right now, stransd doesn't expect us to write a response back to an RPC */
	if (stmp->sm_origin == S_TRANS_KNFSD) {
		wbuf = s_trans_set_wbuf(stmp, NULL, 0 );
		if (!wbuf)
			err =  -1;
		*wbufp = wbuf;
		*wbuflen = sizeof(*stmp);
	}
out:
	syslog(LOG_INFO, "%s: ADD_STATE returns %d\n", __FUNCTION__, err);
	return err;
}

int
s_trans_remove_state(struct s_trans_msg *stmp, char **wbufp, int *wbuflen)
{
	int err = -1;
	char dirname[HEXDIR_LEN + 1], *wbuf = NULL;
	uint32_t *p;

	if (stmp->sm_len != HEXDIR_LEN_PADDED) {
		syslog(LOG_INFO, "%s: Invalid message data length\n", __FUNCTION__);
		stmp->sm_status = S_TRANS_STATUS_INVALIDMSG;
		goto out;
	}

	/* skip past header; followed by a hexdir string */
	p = (uint32_t *)stmp;
	p += 4;
	memcpy(dirname, (char *)p, HEXDIR_LEN);
	dirname[HEXDIR_LEN] = '\0';
	syslog(LOG_INFO, "%s: REMOVE_STATE dir saved as %s\n", __FUNCTION__, dirname);

	if (rm_rf(dirname)) {
		syslog(LOG_INFO, "%s: failed to rmdir for REMOVE STATE\n", __FUNCTION__);
		goto out;
	}
	err = 0;

	/* XXX: right now, stransd doesn't expect us to write a response back to an RPC */
	if (stmp->sm_origin == S_TRANS_KNFSD) {
		wbuf = s_trans_set_wbuf(stmp, NULL, 0 );
		if (!wbuf)
			err =  -1;
		*wbufp = wbuf;
		*wbuflen = sizeof(*stmp);
	}
out:
	syslog(LOG_INFO, "%s: REMOVE_STATE returns %d\n", __FUNCTION__, err);
	return err;
}

/*
 * Process data read from pipe (stored in stmp).
 * Construct any reply data
 * Return prepared write data in wbuf. wbuflen = wbuf size in bytes.
 *
 * On success, returns 0.
 * On error, returns -1.  The original s_trans_msg passed-in will have 
 *   had its sm_status set to S_TRANS_STATUS_FAILED and the caller can just
 *   pass that original message to s_trans_error() to make the downcall.
 */
int
do_state_trans(struct s_trans_msg *stmp, char **wbufp, int *wbuflen)
{
	int err = -1;

	stmp->sm_status = S_TRANS_STATUS_FAILED;

	switch (stmp->sm_type) {
	case S_TRANS_LOAD_REBOOT_STATE:
		/* 
		 * sm_len should be 4, input (read) data should be server boot time.
		 * construct clientid list for output (write) data
		 */
		err = s_trans_load_state(stmp, wbufp, wbuflen);
		break;

	case S_TRANS_ADD_STATE:
		/*
		 * sm_len should be HEXDIR_LEN + 4 for XDR length.
		 * data should contain 1 clientid directory name and a u32 client IP addr.
		 * no output (write) data.
		 */
		syslog(LOG_INFO, "%s: ADD_STATE, stmp->sm_len %d\n", __FUNCTION__, (int)stmp->sm_len);
		err = s_trans_add_state(stmp, wbufp, wbuflen);
		break;
	case S_TRANS_REMOVE_STATE:
		/*
		 * sm_len should be HEXDIR_LEN 
		 * data should contain 1 clientid directory name
		 * no output (write) data.
		 */
		syslog(LOG_INFO, "%s: REMOVE_STATE, stmp->sm_len %d\n", __FUNCTION__, (int)stmp->sm_len);
		err = s_trans_remove_state(stmp, wbufp, wbuflen);
		break;
	default:
		syslog(LOG_INFO, "%s: Invalid type: %d\n", __FUNCTION__, stmp->sm_type);
		stmp->sm_status = S_TRANS_STATUS_INVALIDMSG;
	}

	/* for now, if an op worked, its success is set by set_wbuf..(); if it didn't work, its failure
	 * is registered by the caller of do_state_trans() calling s_trans_error().
	 */
	return err;
}

/*
 * Read/Write arbitrary length data through rpc_pipe_fs.
 */
static void
s_trans_io(void)
{
	struct s_trans_msg *stmp, stm;
	char *rbuf, *wbuf;
	uint32_t *p;
	int wbuflen, wrote;

	syslog(LOG_INFO, "s_trans_io start..\n");

	/* Read in the data header. */
	memset(&stm, 0, sizeof(struct s_trans_msg));
	if (atomicio(read, s_trans_pipe_fd, &stm, sizeof(stm)) != sizeof(stm)) {
		syslog(LOG_INFO, "s_trans_io: read failed. %s\n",
			strerror(errno));
		if (errno == EPIPE)
			return;
		goto out;
	}

	/* Prepare read buffer.  XXX need better bound on stm.sm_len */
	syslog(LOG_INFO, "s_trans_io: read stm.sm_len. %d\n", stm.sm_len);
	if (stm.sm_len > 1024) {
		syslog(LOG_INFO, "s_trans_io: stm.sm_len OUT OF BOUNDS.\n");
		goto out;
	}
	if ((rbuf = calloc(1, sizeof(*stmp) + stm.sm_len)) == NULL) {
		syslog(LOG_INFO, "s_trans_io: calloc failed. %s\n",
			strerror(errno));
		goto out;
	}
	stmp = (struct s_trans_msg *)rbuf;
	stmp->sm_status = stm.sm_status;
	stmp->sm_type = stm.sm_type;
	stmp->sm_origin = stm.sm_origin;
	stmp->sm_len = stm.sm_len;

	if (stm.sm_len == 0)
		goto process;

	/* Read in the payload. */
	p = (uint32_t *)rbuf;
	p += 4;
	if (atomicio(read, s_trans_pipe_fd, (char *)p, stmp->sm_len) != stmp->sm_len) {
		syslog(LOG_INFO, "s_trans_io: read failed. %s\n",
			strerror(errno));
		if (errno == EPIPE)
			return;
		goto out;
	}
	syslog(LOG_INFO, "s_trans_io: read 2 stmp->sm_len. %d\n", stmp->sm_len);

process:
	/* Process the request.  do_state_trans() calloc's the wbuf free'd below */
	if (do_state_trans(stmp, &wbuf, &wbuflen)) {
		s_trans_error(stmp);
		return;
	}
	free(rbuf);

	/* Write any buffered reply data. */
	syslog(LOG_INFO, "s_trans_io: write wbuflen %d\n", wbuflen);
	if ((wrote = atomicio(write, s_trans_pipe_fd, wbuf, wbuflen)) != wbuflen) {
		syslog(LOG_INFO, "s_trans_io: write of %d bytes failed (wrote %d bytes). %s", wbuflen, wrote, strerror(errno));
		stm.sm_status = S_TRANS_STATUS_FAILED;
		s_trans_error(&stm);
	}
	free(wbuf);
	syslog(LOG_INFO, "s_trans_io ends.\n");
out:
	return;
}

/*
 * pause()'s signal handler needs to return in order to correctly resume
 * control, so the signal handler can't contain the pause() call.  we save
 * errno because after a signal errno is set to EINTR.
 */
void
attempt_open(int signal)
{
	s_trans_pipe_fd = open(s_trans_pipe_path, O_RDWR, 0);
	if (s_trans_pipe_fd == -1)
		saved_errno = errno;
}

static void
prepare_dnotify(int fd, struct sigaction *act)
{
	act->sa_handler = attempt_open;
	sigemptyset(&act->sa_mask);
	sigaction(SIGUSR2, act, NULL);
	fcntl(fd, F_SETSIG, SIGUSR2);
	fcntl(fd, F_NOTIFY, DN_CREATE | DN_DELETE | DN_MULTISHOT);
}

static void
cleanup_dnotify(int fd, struct sigaction *act)
{
	fcntl(fd, F_SETSIG, 0);
	fcntl(fd, F_NOTIFY, 0);
	act->sa_handler = SIG_DFL;
	sigaction(SIGUSR2, act, NULL);
}

/* returns 0 on success, -1 on error */
static int
s_trans_init(void)
{
	static struct sigaction initact;
	int err = -1;

	if (nfsd_pipefs_fd != -1)
		goto reattempt;

	/* open recovery dir ..../rpc_pipefs/nfsd/ */
	if ((nfsd_pipefs_fd = open(nfsd_pipefs_dir, O_RDONLY, 0)) == -1) {
		syslog(LOG_INFO, "ERROR: open of nfsd pipefs dir '%s' failed. %s\n", nfsd_pipefs_dir, strerror(errno));
		goto out;
	}
	syslog(LOG_INFO, "Opened nfsd pipefs dir '%s'.\n", nfsd_pipefs_dir);

	attempt_open(0);
reattempt:
	if (s_trans_pipe_fd == -1) {
		switch (saved_errno) {
		case ENOENT:
			if (initact.sa_handler == NULL)
				prepare_dnotify(nfsd_pipefs_fd, &initact);
			pause();
			goto reattempt;
		default:
			close(nfsd_pipefs_fd);
			nfsd_pipefs_fd = -1;
			goto out;
		}
	}
	if (initact.sa_handler != NULL)
		cleanup_dnotify(nfsd_pipefs_fd, &initact);

	syslog(LOG_INFO, "Opened s_trans pipe '%s'.\n", s_trans_pipe_path);
	err = 0;
out:
	return err;
}

/* XXX new header, move back to _svc.c */
int
s_trans_svc_init(void)
{
	register SVCXPRT *transp;
	int err = 1;

	pmap_unset(S_TRANS_PROG, S_TRANS_VERS);

	transp = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transp == NULL) {
		fprintf (stderr, "%s: %s", __FUNCTION__, "cannot create tcp service.");
		goto out;
	}
	if (!svc_register(transp, S_TRANS_PROG, S_TRANS_VERS, s_trans_prog_1, IPPROTO_TCP)) {
		fprintf (stderr, "%s: %s", __FUNCTION__, "unable to register (S_TRANS_PROG, S_TRANS_VERS, tcp).");
		goto out;
	}

	err = 0;
out:
	return err;
}

void
s_trans_run(int pipefd)
{
	int err;
	fd_set rfds;

	for (;;) {
		/* svc_fdset can change at any time .. */
		rfds = svc_fdset;
		FD_SET(pipefd, &rfds);

		err = select(FD_SETSIZE, &rfds, NULL, NULL, NULL);
		if (err == -1) {
			syslog(LOG_INFO, "An error occurred with select() (%s).\n", strerror(errno));
			return;
		}
		/* an upcall or an RPC? */
		if (FD_ISSET(pipefd, &rfds))
			s_trans_io();
		else
			svc_getreqset(&rfds);
	}
}

void
sig_die(int signal)
{
	syslog(LOG_INFO, "Exiting on signal %d\n", signal);
	exit(1);
}

static void
usage()
{
	printf("Usage: %s <NFSv4 server recovery directory> <NFSv4 server rpc_pipefs mount point>\n", program);
}

int
main(int argc, char **argv)
{
	uid_t uid = 0;

	program = basename(argv[0]);

	if ((uid = getuid()) != 0) {
		fprintf(stderr, "%s: uid %d : must run as root\n", program, uid);
		return EPERM;
	}

	if (argc != 3) {
		usage();
		printf("Using default v4 recovery dir '%s' and pipefsdir '%s'\n",
			nfsd_recovery_dir, nfsd_pipefs_dir);
	} else {
		nfsd_recovery_dir = argv[1];
		nfsd_pipefs_dir = argv[2];
		if (strlen(nfsd_recovery_dir) > PATH_MAX ||
				strlen(nfsd_pipefs_dir) > PATH_MAX) {
			syslog(LOG_INFO, "daemon: path name too long\n");
			return EINVAL;
		}
	}

	signal(SIGINT, sig_die);
	signal(SIGTERM, sig_die);
	errno = 0;

	if (chdir(nfsd_recovery_dir) < 0 ) {
		syslog(LOG_INFO, "chdir to daemon dir: %s failed. %s\n",
			nfsd_recovery_dir, strerror(errno));
		return errno;
	}
	syslog(LOG_INFO, "%s starting (using recovery dir %s)..\n", nfsd_recovery_dir, program);

	if (s_trans_init()) {
		syslog(LOG_INFO, "ERROR: %s unable to set up RPC pipe.\n", program);
		return 1;
	}
	if (s_trans_svc_init()) {
		syslog(LOG_INFO, "Unable to start RPC services.\n");
		return 1;
	}

	/* doesn't return */
	s_trans_run(s_trans_pipe_fd);
	return 2;
}
